#include <cstdio>
#include <algorithm>

#include "groupscreen.h"
#include "fetch_group.h"
#include "fetch_max_score_for_song.h"
#include "fetch_max_score_for_player.h"
#include "fetch_needs_update.h"
#include "fonts.h"

GroupScreen::GroupScreen(pqxx::connection &conn, unsigned tournament, unsigned round, unsigned parallel)
	: tournament(tournament), round(round), parallel(parallel), scores_changed(conn, "scores"), conn(conn), valid(false)
{
}

GroupScreen::~GroupScreen()
{
}

bool GroupScreen::check_invalidated()
{
	if (!valid)
		return true;
	if (!scores_changed.get_flag())
		return false;

	bool needs_update;
	conn.perform(FetchNeedsUpdate(last_updated, tournament, round, parallel, &needs_update));

	if (!needs_update)
		scores_changed.reset_flag();
	
	return needs_update;
}

void GroupScreen::draw(unsigned char *buf)
{
	scores_changed.reset_flag();

	Group group;
	conn.perform(FetchGroup(tournament, round, parallel, &group));
	gettimeofday(&last_updated, NULL);

	memset(buf, 0, 800 * 600 * 4);

	// main heading
	char heading[64];
	if (parallel == 0) {
		sprintf(heading, "Round %u", round);
	} else {
		sprintf(heading, "Round %u, Group %u", round, parallel);
	}

	{
		unsigned width = my_draw_text(heading, NULL, 48.0);
		my_draw_text(heading, buf, 48.0, 800/2 - width/2, 60);
	}
	
	// Find out how wide each column has to be. First try unlimited width (ie.
	// long titles for everything); if that gets too long, try again with short
	// titles for chosen songs.
	unsigned width[16], num_scores;
	unsigned max_num_width = my_draw_text("8888", NULL, 22.0);
	unsigned mode;
	for (mode = 0; mode < 2; ++mode) {
		for (unsigned i = 0; i < 16; ++i)
			width[i] = 0;

		for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
			unsigned col = 1;
			width[0] = std::max(width[0], my_draw_text(i->nick, NULL, 18.0));

			for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
				if (j->chosen) {
					width[col] = std::max(width[col], my_draw_text((mode == 0) ? j->song.title : j->song.short_title, NULL, 12.0) + 
							max_num_width + 10);
				} else {		
					width[col] = std::max(width[col], my_draw_text(j->song.short_title, NULL, 12.0));
					width[col] = std::max(width[col], max_num_width);
				}
			}
		}

		num_scores = group.players[0].scores.size();

		width[num_scores + 1] = std::max(my_draw_text("Total", NULL, 12.0), max_num_width);
		width[num_scores + 2] = my_draw_text("Rank", NULL, 12.0);

		// if we're at long titles and that works, don't try the short ones
		if (mode == 0) {
			unsigned sumwidth = 0;
			for (unsigned i = 0; i <= num_scores + 2; ++i)
				sumwidth += width[i] + 20;
			
			if (sumwidth < 800)
				break;
		}
	}

	// make column headings from the first player's songs
	unsigned col = 1;
	unsigned x = 40 + width[0];
	for (std::vector<Score>::const_iterator i = group.players[0].scores.begin(); i != group.players[0].scores.end(); ++i, ++col) {
		if (!i->chosen) {
			unsigned this_width = my_draw_text(i->song.short_title, NULL, 12.0);
			my_draw_text(i->song.short_title, buf, 12.0, x + width[col] / 2 - this_width / 2, 100);
		}
		x += width[col] + 20;
	}

	my_draw_text("Total", buf, 12.0, x + width[num_scores + 1] / 2 - my_draw_text("Total", NULL, 12.0) / 2, 100);
	x += width[num_scores + 1] + 20;
	my_draw_text("Rank", buf, 12.0, x + width[num_scores + 2] / 2 - my_draw_text("Rank", NULL, 12.0) / 2, 100);
	
	// show all the players and the scores
	unsigned y = 140;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		my_draw_text(i->nick, buf, 18.0, 20, y);

		unsigned x = 40 + width[0];

		unsigned col = 1;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
			char text[16] = "";
			if (j->score != -1) {
				sprintf(text, "%u", j->score);
			}
	
			unsigned this_width = my_draw_text(text, NULL, 22.0);
			if (j->chosen) {
				my_draw_text(text, buf, 22.0, x + max_num_width - this_width, y);
				my_draw_text((mode == 0) ? j->song.title : j->song.short_title, buf, 12.0, x + max_num_width + 10, y);
			} else {
				my_draw_text(text, buf, 22.0, x + width[col] / 2 - this_width / 2, y);
			}
			x += width[col] + 20;
		}

		// draw total
		{
			char text[16];
			sprintf(text, "%u", i->total);
			
			unsigned this_width = my_draw_text(text, NULL, 22.0);
			my_draw_text(text, buf, 22.0, x + width[num_scores + 1] / 2 - this_width / 2, y);
			x += width[num_scores + 1] + 20;
		}

		y += 40;
	}
	
	/*
	 * Approximate (but probably working quite well in practice) heuristic
	 * for finding the min and max rank of a player works as follows:
	 *
	 * First of all, find out, for each player in the group, what the
	 * maximum remaining score possibly can be (the minimum score is of
	 * course identical to the player's current total). For a random song,
	 * this is of course 1000 * (maximum feet rating) (but of course, that
	 * depends on whether we can play single or double! for now, assume
	 * double is okay, but this logic will be deferred to FetchMaxScore
	 * anyhow); for a random song, we simply pick the highest-ranking song
	 * we can find, EXCEPT those the player has chosen earlier AND the
	 * random songs this round, AND all random songs from elimination rounds
	 * (ie. rounds with only one group). (Phew!) This doesn't solve problems
	 * we'd face with more than one chosen song, but it should be good enough.
	 *
	 * After we've found the max and min scores for all players, it's a simple
	 * matter of sorting; the best attainable rank for player X is obtained if 
	 * X gets max score and all others get min score, the worst attainable rank
	 * is obtained if X gets min score and all others get max score.
	 *
	 * This is a bit SQL-heavy, but heck...
	 */
	std::vector<unsigned> max_score, min_score;
	for (std::vector<Player>::const_iterator i = group.players.begin(); i != group.players.end(); ++i) {
		unsigned min_score_tp = 0, max_score_tp = 0;
		for (std::vector<Score>::const_iterator j = i->scores.begin(); j != i->scores.end(); ++j, ++col) {
			if (j->score != -1) {
				// already given
				min_score_tp += j->score;
				max_score_tp += j->score;
			} else {
				unsigned max_score_this_song;
				if (j->song.id != -1) {
					// random song, or we know what song the player picked
					conn.perform(FetchMaxScoreForSong(tournament, j->song.id, &max_score_this_song));
				} else {
					conn.perform(FetchMaxScoreForPlayer(tournament, i->id, round, &max_score_this_song));
				}
				max_score_tp += max_score_this_song;
			}
		}
		max_score.push_back(max_score_tp);
		min_score.push_back(min_score_tp);
	}

	// now finally find min and max rank, and draw it all
	y = 140;
	for (unsigned i = 0; i < group.players.size(); ++i) {
		unsigned best_rank = 1, worst_rank = 1;
		for (unsigned j = 0; j < group.players.size(); ++j) {
			if (i == j)
				continue;

			if (max_score[i] < min_score[j])
				++best_rank;
			if (min_score[i] <= max_score[j])
				++worst_rank;
		}

		char text[16];
		if (best_rank == worst_rank)
			std::sprintf(text, "%u", best_rank);
		else
			std::sprintf(text, "%u-%u", best_rank, worst_rank);
		
		unsigned this_width = my_draw_text(text, NULL, 22.0);
		my_draw_text(text, buf, 22.0, x + width[num_scores + 2] / 2 - this_width / 2, y);

		y += 40;
	}
		
	valid = true;
}

